#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <typeinfo.h>
#include "../include/sprite.h"
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>				
Sprite  *SpriteAppWin::LastSprite() {
  Sprite  *ds = first;
  if(ds == NULL)
    return(NULL);
  else
    while (ds->next != NULL) {
      ds=ds->next;
    }
  return(ds);
}

void SpriteAppWin::AddSprite(Sprite  *ds) {
  Sprite  *last;
  if (first == NULL) {
    first = ds;
    ds->prev = NULL;
  }
  else
    {
      last = LastSprite();
      ds->prev = last;
      last->next = ds;
    }
  ds->next = NULL;
}

void SpriteAppWin::DeleteSprite(Sprite  *ds) {
  Sprite  *last,*next;
  if(ds->prev == NULL) {
    first = ds->next;
    if(ds->next != NULL)
      (ds->next)->prev = NULL;
  }
  else
    {
      last = ds->prev;
      last->next = ds->next;
    }
  next = ds->next;
  if(next != NULL) {next->prev = last;}
}

void SpriteAppWin::PlaceSpriteBehind(Sprite  *ds1,Sprite  *ds2) {
  if(ds1 != ds2) {
    Sprite  *p;
    DeleteSprite(ds1);
    if(ds2 == NULL) {AddSprite(ds1);}
    else
      {
	p = ds2->prev;
	ds1->prev = p;
	ds1->next = ds2;
	ds2->prev = ds1;
	if(p != NULL) {p->next = ds1;} else {first = ds1;}
      }
  }
}
SpriteAppWin::SpriteAppWin(Display *dpy,char *_title,int _cx,int _cy) {
  XGCValues gcv;
  XTextProperty tp;
  display = dpy;
  screen = DefaultScreen(dpy);
  cx = _cx;
  cy = _cy;
  title = _title;
  first = NULL;
  _2bnuked = NULL;
  quitting = 0;
  cmap = DefaultColormap(dpy,screen);
  root = DefaultRootWindow(dpy);
  wnd = XCreateSimpleWindow(dpy,root,0,0,_cx,_cy,1,
			    BlackPixel(dpy,screen),BlackPixel(dpy,screen));
  Backpxm = XCreatePixmap(dpy,root,_cx,_cy,DefaultDepth(dpy,screen));
  Bufpxm = XCreatePixmap(dpy,root,_cx,_cy,DefaultDepth(dpy,screen));
  XStringListToTextProperty(&title,1,&tp);
  XSetWMName(dpy,wnd,&tp);
  XStoreName(dpy,wnd,_title);
  XMapRaised(dpy,wnd);
  XSelectInput(display,wnd,ExposureMask | KeyPressMask | KeyReleaseMask);
  gcv.function = GXcopy;
  gcv.graphics_exposures = False;
  bufgc = XCreateGC(display,root,GCFunction |  GCGraphicsExposures ,&gcv);
}

SpriteAppWin::~SpriteAppWin() {
  Quit();
  printf("Destroying window....\n");
  XDestroyWindow(display,wnd);
  XFreeGC(display,bufgc);
}

void SpriteAppWin::Quit() {
  quitting = 1;
}

void SpriteAppWin::_SpriteThread(void *) {
  
}

void SpriteAppWin::Run() {
  XEvent evt;
  while(!quitting) {
    XNextEvent(display,&evt);
    if(evt.xany.window != wnd) continue;
    if((evt.type == Expose)) {
      XCopyArea(display,Bufpxm,wnd,bufgc,0,0,cx,cy,0,0);
      break;
    }
  }
  while(!quitting) {
    while(QLength(display) > 0) {
      XNextEvent(display,&evt);
      if(evt.xany.window != wnd) continue;
      ProcessEvent(&evt);
    }
    _SpriteLoop();
    XCopyArea(display,Bufpxm,wnd,bufgc,0,0,cx,cy,0,0);
    XSync(display,False);
  }
}

int SpriteAppWin::LoadPixmap(char **data,Pixmap *pm_return,
			     Pixmap *mask_return) {
  XpmAttributes attr;
  attr.valuemask = XpmColormap | XpmCloseness;
  attr.colormap = cmap;
  attr.closeness = 65535;
  return XpmCreatePixmapFromData(theDisplay(),
				 root,data,pm_return,mask_return,&attr);
}

//Main sprite loop.
void SpriteAppWin::_SpriteLoop() {
  struct timeval t;
  clock_t c2;
  clock_t c1 = clock();
  XCopyArea(display,Backpxm,Bufpxm,bufgc,0,0,cx,cy,0,0);
  //Loop through all the available sprites, move, render, and check for
  //nuking (deferred deletion).
  Sprite *s;
  for(s = first;s != NULL;s = s->next) {
    try {s->MoveSprite(s->x + s->vx,s->y + s->vy);
    if(s->AutoHitTest()) {
      for(Sprite *s2 = first;s2 != NULL;s2 = s2->next) {
	if(s2 != s) {
	  if(s->SpriteHitTest(s2)) s->OnHitTest(s2);
	}
      }
    }
    } catch(...) {}
    if(_2bnuked != NULL) {
      delete _2bnuked;
      _2bnuked = NULL;
    }
  }
  for(s = first;s != NULL;s = s->next) {
    s->RenderSprite((Drawable)Bufpxm);
  }
  c2=clock(); if((c2 - c1) < 2) {
    t.tv_sec = 0;
    t.tv_usec = 2000000 / CLOCKS_PER_SEC;
    select(0,NULL,NULL,NULL,&t);
  }
}

void SpriteAppWin::ProcessEvent(XEvent *evt) {
  switch(evt->type) {
  case Expose:
    XCopyArea(display,Bufpxm,wnd,bufgc,0,0,cx,cy,0,0);
  }
  
}

void Sprite::CreateSpriteBitmaps(char **data) {
  XpmAttributes attr;
  attr.valuemask = XpmColormap | XpmCloseness;
  attr.colormap = host->cmap;
  attr.closeness = 40000;
  if(pxm != 0) {
    XFreePixmap(host->display,pxm);
  }
  if(maskpxm != 0) {
    XFreePixmap(host->display,maskpxm);
  }
  XpmCreatePixmapFromData(host->display,host->root,data,&pxm,
			  &maskpxm,
			  &attr);
}



void Sprite::CreateSpriteGC() {
  XGCValues gcv;
  gcv.clip_mask = maskpxm;
  gcv.function = GXcopy;
  gcv.graphics_exposures = False;
  
  
  sprgc = XCreateGC(host->display,
		    host->root,GCClipMask |
		    GCGraphicsExposures |
		    GCFunction,&gcv);
}

void Sprite::RenderSprite(Drawable d) {
  XGCValues gcv;
  gcv.clip_x_origin = (int)x;
  gcv.clip_y_origin = (int)y;
  XChangeGC(host->display,sprgc,GCClipXOrigin | GCClipYOrigin,&gcv);
  XCopyArea(host->display,pxm,d,sprgc,0,0,cx,cy,(int)x,(int)y);
}

Sprite::Sprite(SpriteAppWin *sw,float _x,float _y,char **data) {
  host = sw;
  x = _x;
  y = _y;
  vx = 0; vy = 0;
  pxm = 0;
  maskpxm = 0;
  unsigned long dummy1;
  int dummy2;
  CreateSpriteBitmaps(data);
  XGetGeometry(sw->display,pxm,&dummy1,&dummy2,&dummy2,
	       &cx,&cy,&dummy2,&dummy2);
  CreateSpriteGC();
  host->AddSprite(this);
  auto_hittest = False;
}

Sprite::Sprite(SpriteAppWin *sw,float _x,float _y,Pixmap p,Pixmap mp) {
  host = sw;
  x = _x;
  y = _y;
  vx = 0; vy = 0;
  pxm = 0;
  maskpxm = 0;
  unsigned long dummy1;
  int dummy2;
  pxm = p; maskpxm = mp;
  if(p != 0) {
    XGetGeometry(sw->display,pxm,&dummy1,&dummy2,&dummy2,
		 &cx,&cy,&dummy2,&dummy2);
  }
  CreateSpriteGC();
  host->AddSprite(this);
  auto_hittest = False;
}
Sprite::~Sprite() {
  XFreeGC(host->display,sprgc);
  //XFreePixmap(host->display,pxm);
  //XFreePixmap(host->display,maskpxm);
  host->DeleteSprite(this);
}

void Sprite::MoveSprite(float newx,float newy) {
  x = newx; y = newy;
}

void Sprite::SetSpriteMotion(float _vx,float _vy) {
  vx = _vx; vy = _vy;
}


//Load new Xpm data.
void Sprite::ChangeSpriteShape(char **data) {
  XGCValues gcv;
  unsigned long dummy1;
  int dummy2;
  CreateSpriteBitmaps(data);
  XGetGeometry(host->display,pxm,&dummy1,&dummy2,
	       &dummy2,&cx,&cy,&dummy2,&dummy2);
  gcv.clip_mask = maskpxm;
  XChangeGC(host->display,sprgc,GCClipMask,&gcv);
};

void Sprite::ChangeSpriteShape(Pixmap p,Pixmap mp) {
  XGCValues gcv;
  unsigned long dummy1;
  int dummy2;
  pxm = p; maskpxm = mp;
  XGetGeometry(host->display,pxm,&dummy1,&dummy2,
	       &dummy2,&cx,&cy,&dummy2,&dummy2);
  gcv.clip_mask = maskpxm;
  XChangeGC(host->display,sprgc,GCClipMask,&gcv);
}  

//Do some simple collision checking by comparing the rectangles of
//the source and destination sprites for overlap.
int Sprite::SpriteHitTest(Sprite *s) {
  int step1,step2; //Intermediate x and y proximity values.
  //Step 1: Check for horizontal proximity.
  if(x <= s->XPos()) {
    if(s->XPos() - x < cx) step1 = True; else step1 = False;
  }
  else {
    if(s->XPos() + s->XSize() > x) step1 = True; else step1 = False;
  }
  //Step 2: Check for vertical proximity.
  if(y <= s->YPos()) {
    if(s->YPos() - y < cy) step2 = True; else step2 = False;
  }
  else {
    if(s->YPos() + s->YSize() > y) step2 = True; else step2 = False;
  }
  //Return true if we are overlapping horizontally AND vertically.
  return(step1 && step2);
}

//Default, do-nothing procedure for collision notification.
void Sprite::OnHitTest(Sprite *spr) {
}

//Nuke. Replaces the Win32 delete this functionality. DO NOT use delete this
//in a MoveSprite function under unix!
void Sprite::Nuke() {
  host->_2bnuked = this;
}

































